using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;

using GRV11; // Requires GRV11.cs or a "reference" to GRV11DLL.dll

using System.Drawing;
using System.Drawing.Imaging;

public class CCoordinate
{
    protected double FX = 0D;
    protected double FY = 0D;
    protected double FZ = 0D;

    public delegate void OnChangedHandler(CCoordinate ACoordinate, double AOldX, double AOldY, double AOldZ);

    public event OnChangedHandler OnAssigned;

    protected virtual void RaiseOnAssigned(double AOldX, double AOldY, double AOldZ)
    {
        if (OnAssigned != null)
        {
            OnAssigned(this, AOldX, AOldY, AOldZ);
        }
    }

    public CCoordinate()
    {
        FX = 0D;
        FY = 0D;
        FZ = 0D;
    }

    public CCoordinate(double AX, double AY, double AZ)
    {
        FX = AX;
        FY = AY;
        FZ = AZ;
    }

    public double X
    {
        get
        {
            return FX;
        }

        set
        {
            Assign(value, FY, FZ);
        }
    }

    public float Xf
    {
        get
        {
            return (float)(FX);
        }

        set
        {
            Assign((double)(value), FY, FZ);
        }
    }

    public double Y
    {
        get
        {
            return FY;
        }

        set
        {
            Assign(FX, value, FZ);
        }
    }

    public float Yf
    {
        get
        {
            return (float)(FY);
        }

        set
        {
            Assign(FX, (double)(value), FZ);
        }
    }

    public double Z
    {
        get
        {
            return FZ;
        }

        set
        {
            Assign(FX, FY, value);
        }
    }

    public float Zf
    {
        get
        {
            return (float)(FZ);
        }

        set
        {
            Assign(FX, FY, (double)(value));
        }
    }

    public bool Assign(double AX, double AY, double AZ)
    {
        double LX = FX;
        double LY = FY;
        double LZ = FZ;

        FX = AX;
        FY = AY;
        FZ = AZ;

        RaiseOnAssigned(LX, LY, LZ);

        return true;
    }

    public bool Assign(CCoordinate ACoordinate)
    {
        bool LResult = false;

        if (ACoordinate != null)
        {
            LResult = Assign(ACoordinate.X, ACoordinate.Y, ACoordinate.Z);
        }

        return LResult;
    }

    public bool Equals(double AX, double AY, double AZ)
    {
        return ((FX == AX) && (FY == AY) && (FZ == AZ));
    }

    public bool Equals(CCoordinate ACoordinate)
    {
        bool LResult = false;

        if (ACoordinate != null)
        {
            LResult = Equals(ACoordinate.X, ACoordinate.Y, ACoordinate.Z);
        }

        return LResult;
    }

    public bool Rotate(double AX, double AY, double AZ, bool APositive)
    {
        double LTempX = FX;
        double LTempY = FY;
        double LTempZ = FZ;

        if (APositive)
        {
            // rotation around Z axis
            FX = LTempX * Math.Cos(AZ) - LTempY * Math.Sin(AZ);
            FY = LTempX * Math.Sin(AZ) + LTempY * Math.Cos(AZ);
            FZ = LTempZ;

            // save temporarily
            LTempX = FX;
            LTempY = FY;
            LTempZ = FZ;

            // rotation around Y axis
            FX = LTempZ * Math.Sin(AY) + LTempX * Math.Cos(AY);
            FY = LTempY;
            FZ = LTempZ * Math.Cos(AY) - LTempX * Math.Sin(AY);

            // save temporarily
            LTempX = FX;
            LTempY = FY;
            LTempZ = FZ;

            // rotation around X axis
            FX = LTempX;
            FY = LTempY * Math.Cos(AX) - LTempZ * Math.Sin(AX);
            FZ = LTempY * Math.Sin(AX) + LTempZ * Math.Cos(AX);
        }
        else
        {
            // rotation around X axis
            FX = LTempX;
            FY = LTempY * Math.Cos(- AX) - LTempZ * Math.Sin(- AX);
            FZ = LTempY * Math.Sin(- AX) + LTempZ * Math.Cos(- AX);

            // save temporarily
            LTempX = FX;
            LTempY = FY;
            LTempZ = FZ;

            // rotation around Y axis
            FX = LTempZ * Math.Sin(- AY) + LTempX * Math.Cos(- AY);
            FY = LTempY;
            FZ = LTempZ * Math.Cos(- AY) - LTempX * Math.Sin(- AY);

            // save temporarily
            LTempX = FX;
            LTempY = FY;
            LTempZ = FZ;

            // rotation around Z axis
            FX = LTempX * Math.Cos(- AZ) - LTempY * Math.Sin(- AZ);
            FY = LTempX * Math.Sin(- AZ) + LTempY * Math.Cos(- AZ);
            FZ = LTempZ;
        }

        return true;
    }

    public bool Rotate(CCoordinate ACoordinate, bool APositive)
    {
        bool LResult = false;

        if (ACoordinate != null)
        {
            LResult = Rotate(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, APositive);
        }

        return LResult;
    }

    public bool Scale(double AX, double AY, double AZ, double AMultiplier, bool APositive)
    {
        bool LResult = false;

        if (APositive)
        {
            FX *= AX * AMultiplier;
            FY *= AY * AMultiplier;
            FZ *= AZ * AMultiplier;

            LResult = true;
        }
        else
        {
            if ((AMultiplier != 0D) && (AX != 0D) && (AY != 0D) && (AZ != 0D))
            {
                FX /= AX * AMultiplier;
                FY /= AY * AMultiplier;
                FZ /= AZ * AMultiplier;

                LResult = true;
            }
        }

        return LResult;
    }

    public bool Scale(CCoordinate ACoordinate, double AMultiplier, bool APositive)
    {
        bool LResult = false;

        if (ACoordinate != null)
        {
            LResult = Scale(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, AMultiplier, APositive);
        }

        return LResult;
    }

    public bool Translate(double AX, double AY, double AZ, bool APositive)
    {
        if (APositive)
        {
            FX += AX;
            FY += AY;
            FZ += AZ;
        }
        else
        {
            FX -= AX;
            FY -= AY;
            FZ -= AZ;
        }

        return true;
    }

    public bool Translate(CCoordinate ACoordinate, bool APositive)
    {
        bool LResult = false;

        if (ACoordinate != null)
        {
            LResult = Translate(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, APositive);
        }

        return LResult;
    }

    public static double DegToRad(double AAngle)
    {
        return (AAngle / 180D) * Math.PI;
    }

    public static double RadToDeg(double AAngle)
    {
        return (AAngle / Math.PI) * 180D;
    }
}

public class CCube
{
    protected CCoordinate[,] FMatrix;
    protected CCoordinate[]  FPoints;

    protected CCoordinate FPos;
    protected CCoordinate FRot;
    protected CCoordinate FScale;

    protected double FScaleMultiplier = 50D;

    protected void OnAssignedPos(CCoordinate ACoordinate, double AOldX, double AOldY, double AOldZ)
    {
        if (!(FPos.Equals(AOldX, AOldY, AOldZ)))
        {
            foreach (CCoordinate LIndex in FPoints)
            {
                LIndex.Translate(AOldX, AOldY, AOldZ, false);

                LIndex.Translate(FPos, true);
            }
        }
    }

    protected void OnAssignedRot(CCoordinate ACoordinate, double AOldX, double AOldY, double AOldZ)
    {
//        if (!(FRot.Equals(AOldX, AOldY, AOldZ)))
//        {
            foreach (CCoordinate LIndex in FPoints)
            {
                LIndex.Translate(FPos, false);
                LIndex.Rotate(AOldX, AOldY, AOldZ, false);

                LIndex.Rotate(FRot, true);
                LIndex.Translate(FPos, true);
            }
//        }
    }

    protected void OnAssignedScale(CCoordinate ACoordinate, double AOldX, double AOldY, double AOldZ)
    {
        if (!(FScale.Equals(AOldX, AOldY, AOldZ)))
        {
            foreach (CCoordinate LIndex in FPoints)
            {
                LIndex.Translate(FPos, false);
                LIndex.Rotate(FRot, false);
                LIndex.Scale(AOldX, AOldY, AOldZ, FScaleMultiplier, false);

                LIndex.Scale(FScale, FScaleMultiplier, true);
                LIndex.Rotate(FRot, true);
                LIndex.Translate(FPos, true);
            }
        }
    }

    public CCube()
    {
        FPoints    = new CCoordinate[8];
        FPoints[0] = new CCoordinate(- 0.5D, - 0.5D, - 0.5D);
        FPoints[1] = new CCoordinate(+ 0.5D, - 0.5D, - 0.5D);
        FPoints[2] = new CCoordinate(- 0.5D, + 0.5D, - 0.5D);
        FPoints[3] = new CCoordinate(- 0.5D, - 0.5D, + 0.5D);
        FPoints[4] = new CCoordinate(+ 0.5D, + 0.5D, - 0.5D);
        FPoints[5] = new CCoordinate(+ 0.5D, - 0.5D, + 0.5D);
        FPoints[6] = new CCoordinate(- 0.5D, + 0.5D, + 0.5D);
        FPoints[7] = new CCoordinate(+ 0.5D, + 0.5D, + 0.5D);

        FMatrix = new CCoordinate[6,4];

        // top
        FMatrix[0,0] = FPoints[2];
        FMatrix[0,1] = FPoints[6];
        FMatrix[0,2] = FPoints[7];
        FMatrix[0,3] = FPoints[4];

        // bottom
        FMatrix[1,0] = FPoints[1];
        FMatrix[1,1] = FPoints[5];
        FMatrix[1,2] = FPoints[3];
        FMatrix[1,3] = FPoints[0];

        // left
        FMatrix[2,0] = FPoints[2];
        FMatrix[2,1] = FPoints[0];
        FMatrix[2,2] = FPoints[3];
        FMatrix[2,3] = FPoints[6];

        // right
        FMatrix[3,0] = FPoints[7];
        FMatrix[3,1] = FPoints[5];
        FMatrix[3,2] = FPoints[1];
        FMatrix[3,3] = FPoints[4];

        // front
        FMatrix[4,0] = FPoints[6];
        FMatrix[4,1] = FPoints[3];
        FMatrix[4,2] = FPoints[5];
        FMatrix[4,3] = FPoints[7];

        // back
        FMatrix[5,0] = FPoints[4];
        FMatrix[5,1] = FPoints[1];
        FMatrix[5,2] = FPoints[0];
        FMatrix[5,3] = FPoints[2];

        FPos   = new CCoordinate(0D, 0D, 0D);
        FRot   = new CCoordinate(0D, 0D, 0D);
        FScale = new CCoordinate(1D, 1D, 1D);

        FPos.OnAssigned   += new CCoordinate.OnChangedHandler(OnAssignedPos);
        FRot.OnAssigned   += new CCoordinate.OnChangedHandler(OnAssignedRot);
        FScale.OnAssigned += new CCoordinate.OnChangedHandler(OnAssignedScale);

        FScaleMultiplier = 1D;
    }

    public double ScaleMultiplier
    {
        get
        {
            return FScaleMultiplier;
        }

        set
        {
            if (value != FScaleMultiplier)
            {
                foreach (CCoordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);
                    LIndex.Rotate(FRot, false);
                    LIndex.Scale(FScale, FScaleMultiplier, false);

                    LIndex.Scale(FScale, value, true);
                    LIndex.Rotate(FRot, true);
                    LIndex.Translate(FPos, true);
                }

                FScaleMultiplier = value;
            }
        }
    }

    public CCoordinate Pos
    {
        get
        {
            return FPos;
        }

        set
        {
            if (!(FPos.Equals(value)))
            {
                foreach (CCoordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);

                    LIndex.Translate(value, true);
                }

                FPos.Assign(value);
            }
        }
    }

    public CCoordinate Rot
    {
        get
        {
            return FRot;
        }

        set
        {
            if (!(FRot.Equals(value)))
            {
                foreach (CCoordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);
                    LIndex.Rotate(FRot, false);

                    LIndex.Rotate(value, true);
                    LIndex.Translate(FPos, true);
                }

                FRot.Assign(value);
            }
        }
    }

    public CCoordinate Scale
    {
        get
        {
            return FScale;
        }

        set
        {
            if (!(FScale.Equals(value)))
            {
                foreach (CCoordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);
                    LIndex.Rotate(FRot, false);
                    LIndex.Scale(FScale, FScaleMultiplier, false);

                    LIndex.Scale(value, FScaleMultiplier, true);
                    LIndex.Rotate(FRot, true);
                    LIndex.Translate(FPos, true);
                }

                FScale.Assign(value);
            }
        }
    }

    public bool IsCollision(GR PGR, double AX, double AY, double AZ, bool ADraw, ref double ANewY)
    {
        bool LResult = false;

        double LXmax = FPoints[0].X;
        double LXmin = FPoints[0].X;
        double LYmax = FPoints[0].Y;
        double LYmin = FPoints[0].Y;
        double LZmax = FPoints[0].Z;
        double LZmin = FPoints[0].Z;

        for (int LIndex = 1; LIndex < FPoints.GetLength(0); LIndex++)
        {
            if (FPoints[LIndex].X < LXmin)
            {
                LXmin = FPoints[LIndex].X;
            }
            if (FPoints[LIndex].X > LXmax)
            {
                LXmax = FPoints[LIndex].X;
            }

            if (FPoints[LIndex].Y < LYmin)
            {
                LYmin = FPoints[LIndex].Y;
            }
            if (FPoints[LIndex].Y > LYmax)
            {
                LYmax = FPoints[LIndex].Y;
            }

            if (FPoints[LIndex].Z < LZmin)
            {
                LZmin = FPoints[LIndex].Z;
            }
            if (FPoints[LIndex].Z > LZmax)
            {
                LZmax = FPoints[LIndex].Z;
            }
        }

        if (ADraw)
        {
            PGR.glPushMatrix();
            PGR.glBegin(GR.GL_QUADS);

            PGR.glColor3f(1f, 0f, 0f);

            PGR.glVertex3f((float)LXmin, (float)LYmax, (float)LZmin);
            PGR.glVertex3f((float)LXmin, (float)LYmax, (float)LZmax);
            PGR.glVertex3f((float)LXmax, (float)LYmax, (float)LZmax);
            PGR.glVertex3f((float)LXmax, (float)LYmax, (float)LZmin);

            PGR.glEnd();
            PGR.glPopMatrix();
        }
        
        LResult = ((LXmin <= AX) && (LXmax >= AX) && (LYmin - 1 <= AY) && (LYmax + 1 >= AY) && (LZmin <= AZ) && (LZmax >= AZ));
        if (LResult)
        {
            ANewY = LYmax;
        }

        return LResult;
    }

    public void DrawCube(GR PGR)
    {
        PGR.glMatrixMode(GR.GL_MODELVIEW);
        PGR.glPushMatrix();
        PGR.glBegin(GR.GL_QUADS);

        for (int LIndexA = 0; LIndexA < FMatrix.GetLength(0); LIndexA++)
        {
            PGR.glColor3f(1f / 6f * (float)(6 - LIndexA), 1f / 6f * (float)(6 - LIndexA), 1f / 6f * (float)(6 - LIndexA));

            for (int LIndexB = 0; LIndexB < FMatrix.GetLength(1); LIndexB++)
            {
                PGR.glVertex3f(FMatrix[LIndexA, LIndexB].Xf, FMatrix[LIndexA, LIndexB].Yf, FMatrix[LIndexA, LIndexB].Zf);
            }
        }

        PGR.glEnd();
        PGR.glPopMatrix();
    }
}

public class Handler1
{
    private Form1 mForm = null;

    private GRVector3 mCameraPosition;
    private GRVector3 mCameraAngle;

    private bool[] mIsPressed = new bool[8];
    private bool   mJump      = false;
    private float  mJumpDir   = 1.0f;

    private DateTime FPressed;
    private bool     FDraw    = false;

    private CCube FLast = null;

    protected CCube[] FCubes;

    public void ClearIsPressed()
    {
        for (int index = 0; index < mIsPressed.Length; index++)
        {
            mIsPressed[index] = false;
        }
    }

    public int GetKeyIndex(Keys k)
    {
        switch (k)
        {
            case Keys.Up :
                return 1;
            case Keys.Down :
                return 2;
            case Keys.Left :
                return 3;
            case Keys.Right :
                return 4;
            case Keys.ControlKey :
                return 5;
            case Keys.Escape :
                return 6;
            case Keys.F1:
                return 7;
            default:
                return 0;
        }
    }

    public bool IsKeyPressed(Keys k)
    {
        return mIsPressed[GetKeyIndex(k)];
    }

    public Handler1(Form1 form, GRControl grControl)
    {
        FPressed = DateTime.Now;

        /*
        this.mGRMesh = new GRMesh();
        this.mGRTexture = new GRTexture();
        */

        mForm = form;

        this.mCameraPosition.Set(0.0f, 25.0f, 0.0f);
        this.mCameraAngle.Set(0.0f, 0.0f, 0.0f);

        ClearIsPressed();

        Random LRandom = new Random();

        FCubes = new CCube[LRandom.Next(100, 150) + 1];
        for (int LIndex = 0; LIndex < FCubes.Length; LIndex++)
        {
            FCubes[LIndex] = new CCube();
            FCubes[LIndex].Scale.X = LRandom.NextDouble() * 2;
            FCubes[LIndex].Scale.Y = LRandom.NextDouble() * 2;
            FCubes[LIndex].Scale.Z = LRandom.NextDouble() * 2;

            FCubes[LIndex].Rot.X = 0D;
            FCubes[LIndex].Rot.Y = 0D;
            FCubes[LIndex].Rot.Z = 0D;

            FCubes[LIndex].Pos.X = LRandom.NextDouble() * 125f;
            FCubes[LIndex].Pos.Y = LRandom.NextDouble() * 125f;
            FCubes[LIndex].Pos.Z = LRandom.NextDouble() * 125f;

            if (LRandom.Next(1, 100) > 50)
            {
                FCubes[LIndex].Pos.X = - FCubes[LIndex].Pos.X;
            }
            if (LRandom.Next(1, 100) > 50)
            {
                FCubes[LIndex].Pos.Y = - FCubes[LIndex].Pos.Y;
            }
            if (LRandom.Next(1, 100) >= 50)
            {
                FCubes[LIndex].Pos.Z = - FCubes[LIndex].Pos.Z;
            }

            FCubes[LIndex].ScaleMultiplier = LRandom.Next(15, 25);
        }

        FCubes[0].Scale.X = 1F;
        FCubes[0].Scale.Y = 1F;
        FCubes[0].Scale.Z = 1F;

        FCubes[0].Pos.X = 0F;
        FCubes[0].Pos.Y = 0F;
        FCubes[0].Pos.Z = 0F;

        FCubes[0].ScaleMultiplier = 100F;

        FLast = FCubes[0];

        Cursor.Hide();
        Cursor.Position = new System.Drawing.Point(mForm.Left + grControl.Left + (grControl.Width / 2),
                                                   mForm.Top + grControl.Top + (grControl.Height / 2));
    }

    public void OpenGLStarted(GRControl grControl)
    {
        /*
        GR gr = grControl.GetGR();
        if (null == gr) return;

        this.mGRMesh = new GRMesh();
        this.mGRMesh.BuildSimpleCube(150.0f);

        this.mGRTexture = new GRTexture();
        this.mGRTexture.CreateCheckerboardTexture(gr, 256, 256);
        */
    }

    public void Paint(object sender, PaintEventArgs e)
    {
        CCube Lowest = FLast;
        foreach (CCube CubeIndex in FCubes)
        {
            if ((CubeIndex.Pos.Y - CubeIndex.Scale.X * CubeIndex.ScaleMultiplier) <
                (Lowest.Pos.Y - Lowest.Scale.X * Lowest.ScaleMultiplier))
            {
                Lowest = CubeIndex;
            }
        }
        if (mCameraPosition.mY < (Lowest.Pos.Y - Lowest.Scale.X * Lowest.ScaleMultiplier) - 100)
        {
            mCameraPosition.mX = FLast.Pos.Xf;
            mCameraPosition.mY = FLast.Pos.Yf + FLast.Scale.Yf * (float)FLast.ScaleMultiplier + 30f;
            mCameraPosition.mZ = FLast.Pos.Zf;
        }

        int FPS = System.Environment.TickCount;

        if (null == sender) return;
        if (false == (sender is GRControl)) return;

        GRControl grControl = (sender as GRControl);
        GR gr = grControl.GetGR();

        Random LRandom = new Random();

        for (int LIndex = 0; LIndex < FCubes.GetLength(0); LIndex++)
        {
            FCubes[LIndex].Rot.X += CCoordinate.DegToRad(LRandom.NextDouble() / 5D);
            FCubes[LIndex].Rot.Y += CCoordinate.DegToRad(LRandom.NextDouble() / 5D);
            FCubes[LIndex].Rot.Z += CCoordinate.DegToRad(LRandom.NextDouble() / 5D);
        }

        if (IsKeyPressed(Keys.F1))
        {
            if (((DateTime.Now.Ticks - FPressed.Ticks) / 10e6) > 1)
            {
                FDraw = !(FDraw);

                FPressed = DateTime.Now;
            }
        }

        // do a JUMP
        if (!(mJump))
        {
            mJump = IsKeyPressed(Keys.ControlKey);
            if (mJump)
            {
                mJumpDir = 2f;
            }
        }
        if (mJump)
        {
            mCameraPosition.mY += mJumpDir;

            mJumpDir -= 0.02f;
        }

        // do a MOVEMENT
        double dX = mCameraPosition.mX;
        double dY = mCameraAngle.mY / 180 * Math.PI;
        double dZ = mCameraPosition.mZ;
        if (IsKeyPressed(Keys.Up))
        {
            dX += 0.5f * Math.Sin(dY);
            dZ -= 0.5f * Math.Cos(dY);
        }
        if (IsKeyPressed(Keys.Down))
        {
            dX -= 0.5f * Math.Sin(dY);
            dZ += 0.5f * Math.Cos(dY);
        }
        if (IsKeyPressed(Keys.Left))
        {
            dX -= 0.5f * Math.Cos(dY);
            dZ -= 0.5f * Math.Sin(dY);
        }
        if (IsKeyPressed(Keys.Right))
        {
            dX += 0.5f * Math.Cos(dY);
            dZ += 0.5f * Math.Sin(dY);
        }
        mCameraPosition.mX = (float)dX;
        mCameraPosition.mZ = (float)dZ;

        int clientWidth = grControl.ClientRectangle.Width;
        int clientHeight = grControl.ClientRectangle.Height;
        if (clientWidth <= 0)
        {
            clientWidth = 1;
        }
        if (clientHeight <= 0)
        {
            clientHeight = 1;
        }

        // Viewport - defines drawing range
        gr.glViewport(0, 0, clientWidth, clientHeight);

        // Basic rendering conditions
        gr.glEnable(GR.GL_DEPTH_TEST);
        gr.glEnable(GR.GL_CULL_FACE);
        gr.glCullFace(GR.GL_BACK);
        /*
        gr.glDepthFunc(GR.GL_LEQUAL);
        gr.glFrontFace(GR.GL_CCW);
         */

        gr.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        gr.glClear(GR.GL_COLOR_BUFFER_BIT | GR.GL_DEPTH_BUFFER_BIT);

        gr.glMatrixMode(GR.GL_PROJECTION);
        gr.glLoadIdentity();

        gr.gluPerspective(60.0f, ((double)(clientWidth) / (double)(clientHeight)), 0.1f, 64000f);

        gr.glMatrixMode(GR.GL_MODELVIEW);
        gr.glLoadIdentity();
        gr.glScalef(1f, 1f, 1f);

        //!!! user movement
        gr.glRotatef(mCameraAngle.mX, 1, 0, 0);
        gr.glRotatef(mCameraAngle.mY, 0, 1, 0);
        gr.glRotatef(mCameraAngle.mZ, 0, 0, 1);

        //        gr.glRotatef(0, 0f, 1f, 0f);
        gr.glTranslatef((-(this.mCameraPosition.mX)), (-(this.mCameraPosition.mY)), (-(this.mCameraPosition.mZ)));
        //!!! user movement

        if (!(FDraw))
        {
            for (int LIndex = 0; LIndex < FCubes.GetLength(0); LIndex++)
            {
                FCubes[LIndex].DrawCube(gr);
            }
        }

        // most simple collision detection
        bool LCollide = false;
        double LY = 0D;
        for (int LIndex = 0; (LIndex < FCubes.GetLength(0)) && !(LCollide); LIndex++)
        {
            LCollide = FCubes[LIndex].IsCollision(gr, mCameraPosition.mX, mCameraPosition.mY - 25, mCameraPosition.mZ, FDraw, ref LY);
            if (LCollide)
            {
                FLast = FCubes[LIndex];
            }
        }

        if (LCollide)
        {
            mCameraPosition.mY = (float)LY + 25f;

            mJumpDir = 0;
            mJump = false;
        }
        else
        {
            mJump = true;
        }

        // do a RESET
        if (IsKeyPressed(Keys.Escape))
        {
            LCollide = true;
            mJump    = false;

            mJumpDir = 0f;

            this.mCameraPosition.Set(0.0f, 25.0f, 0.0f);
            this.mCameraAngle.Set(0.0f, 0.0f, 0.0f);
        }

        gr.wglSwapBuffers(grControl.GetHDC());

        FPS = 1000 / (System.Environment.TickCount - FPS + 1);
        mForm.Text = "FPS=" + FPS.ToString() +
                     " aX=" + mCameraAngle.mX.ToString() +
                     " aY=" + mCameraAngle.mY.ToString() +
                     " pX=" + mCameraPosition.mX.ToString() +
                     " pY=" + mCameraPosition.mY.ToString() +
                     " pZ=" + mCameraPosition.mZ.ToString() +
                     " Ju=" + mJump.ToString() + 
                     " Co=" + LCollide.ToString() +
                     " Cc=" + FCubes.GetLength(0).ToString();
    }

    public void KeyPress(object sender, KeyPressEventArgs e)
    {
        // empty
    }

    public void KeyDown(object sender, KeyEventArgs e)
    {
        mIsPressed[GetKeyIndex(e.KeyCode)] = true;
    }

    public void KeyUp(object sender, KeyEventArgs e)
    {
        mIsPressed[GetKeyIndex(e.KeyCode)] = false;
    }

    public void MouseDown(object sender, MouseEventArgs e)
    {
        // empty
    }

    public void MouseUp(object sender, MouseEventArgs e)
    {
        // empty
    }

    public void MouseMove(object sender, MouseEventArgs e)
    {
        if (null == sender) return;
        if (false == (sender is GRControl)) return;
        GRControl grControl = (sender as GRControl);

        Point lCursorPos = Cursor.Position;
        lCursorPos.X -= mForm.Left + grControl.Left + (grControl.Width / 2);
        lCursorPos.Y -= mForm.Top + grControl.Top + (grControl.Height / 2);

        if ((Math.Abs(lCursorPos.X) > 10) || (Math.Abs(lCursorPos.Y) > 10))
        {
            this.mCameraAngle.mX += 0.05f * (float)(lCursorPos.Y);
            this.mCameraAngle.mY += 0.05f * (float)(lCursorPos.X);

            this.mCameraAngle.mY %= 360;
            if (this.mCameraAngle.mY < 0)
            {
                this.mCameraAngle.mY += 360;
            }

            if (mCameraAngle.mX < -55)
            {
                mCameraAngle.mX = -55;
            }
            else
            {
                if (mCameraAngle.mX > 55)
                {
                    mCameraAngle.mX = 55;
                }
            }

            Cursor.Position = new System.Drawing.Point(mForm.Left + grControl.Left + (grControl.Width / 2),
                                                       mForm.Top + grControl.Top + (grControl.Height / 2));
        }
    }

    public void MouseWheel(object sender, MouseEventArgs e)
    {
        // empty
    }
}